from django.apps import apps
from django.core.management.base import BaseCommand

from djplacedata.models import Place
from djplacedata.utils import get_place_field_from_model, get_values_from_model


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('app_label')
        parser.add_argument('model_name',)

    def handle(self, *args, **options):
        app_label = options['app_label']
        model_name = options['model_name']

        self.model = apps.get_model(app_label, model_name)
        self.field = get_place_field_from_model(self.model)
        self.values = get_values_from_model(self.model)

        for place, values in self.values.results:
            try:
                self.model.objects.update_or_create(
                    **{self.field: place,},
                    defaults=values)
            except Place.DoesNotExist:
                pass



def get_place_field_from_model(model):
    places_meta = getattr(model, 'PlacesMeta', None)
    return getattr(places_meta, 'field')


def get_values_from_model(model):
    places_meta = getattr(model, 'PlacesMeta', None)
    return getattr(places_meta, 'values')



from abc import ABC, abstractmethod

from census import Census
from sodapy import Socrata

from .models import Place
from .settings import CENSUS_API_KEY, SODA_APP_TOKEN


class BaseValue(ABC):
    @property
    @abstractmethod
    def fields(self):
        pass

    @abstractmethod
    def get_value_from_result(self, result):
        pass


class Value(BaseValue):
    def __init__(self, input_field):
        self._input_field = input_field

    @property
    def fields(self):
        return [self._input_field,]

    def get_value_from_result(self, result):
        return result[self._input_field]


class ComputedValue(BaseValue):
    def __init__(self, input_field0,  operator, input_field1):
        self._input_field0 = input_field0
        self._operator = operator
        self._input_field1 = input_field1

    @property
    def fields(self):
        return [self._input_field0, self._input_field1,]

    def get_value_from_result(self, result):
        try:
            value = eval(f'{result[self._input_field0]}{self._operator}{result[self._input_field1]}')
            return value
        except:
            return None


class BaseValues(ABC):
    def __init__(self, **fields):
        self.fields = fields

    @property
    def results(self):
        for place, result in self.query():
            values = {}
            for name, value in self.fields.items():
                values[name] = value.get_value_from_result(result)
            yield place, values

    @property
    def queryable_fields(self):
        values = []
        for field in self.fields.values():
            values += field.fields
        return values

    @abstractmethod
    def query(self):
        pass


_STATE_GEOIDS = [
    '01', '02', '04', '05', '06', '08', '09', '10', '11',
    '12', '13', '15', '16', '17', '18', '19', '20', '21',
    '22', '23', '24', '25', '26', '27', '28', '29', '30',
    '31', '32', '33', '34', '35', '36', '37', '38', '39',
    '40', '41', '42', '44', '45', '46', '47', '48', '49',
    '50', '51', '53', '54', '55', '56',
]


class CensusValues(BaseValues):
    def __init__(self, dataset, year, *args, **kwargs):
        super().__init__(*args, **kwargs)
        census = Census(CENSUS_API_KEY, year=year)
        self.client = getattr(census, dataset)

    def query(self):
        results = [
            (Place.objects.get(geoid=result['state']), result)
            for result in self.client.get(self.queryable_fields, {'for': 'state:*'})
        ]
        results += [
            (Place.objects.get(geoid=result['state'] + result['county']), result)
            for result in self.client.get(self.queryable_fields, {'for': 'county:*'})
        ]
        for state_geoid in _STATE_GEOIDS:
            results += [
                (
                    Place.objects.get(geoid=result['state'] + result['county'] + result['county subdivision']),
                    result
                )
                for result in
                self.client.get(self.queryable_fields, {'for': 'county subdivision:*', 'in': 'state:%s' % state_geoid})
            ]
        return results



class SodaAPIValues(BaseValues):
    def __init__(self, domain, dataset, filters, get_place, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = Socrata(
            domain,
            SODA_APP_TOKEN
        )
        self.dataset = dataset
        self.filters = filters
        self.get_place = get_place

    def query(self):
        results = []
        for data in self.client.get(self.dataset, **self.filters):
            place = self.get_place(data)
            if place is None:
                continue
            results.append((place, data))
        return results



from django.db import models
from django.contrib.gis.db.models import MultiPolygonField


class Place(models.Model):
    geoid = models.CharField(max_length=15, unique=True)
    name = models.CharField(max_length=100)
    geom = MultiPolygonField()

    parent = models.ForeignKey('self', related_name='places', on_delete=models.CASCADE, null=True)

    class Type(models.IntegerChoices):
        state = 0
        county = 1
        city = 2

    type = models.IntegerField(choices=Type.choices)



from django.contrib.gis.geos import MultiPolygon, Polygon
from django.contrib.gis.gdal.datasource import DataSource
from django.core.management.base import BaseCommand

from census_shapefiles import CensusShapefiles

from djplacedata.models import Place
from djplacedata.settings import CENSUS_YEAR, CENSUS_SHAPEFILE_SIMPLIFICATION


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            '--year',
            type=int,
            default=CENSUS_YEAR
        )
        parser.add_argument(
            '--simplification',
            default=CENSUS_SHAPEFILE_SIMPLIFICATION,
        )

    def handle(self, *args, **options):
        year = options['year']
        self.simplification = options['simplification']

        shapefiles = CensusShapefiles(year)

        for shapefile in shapefiles.state.get_shapefiles():
            ds = DataSource(shapefile)
            for feature in ds[0]:
                Place.objects.update_or_create(
                    geoid=feature.get('GEOID'),
                    defaults={
                        'name': feature.get('NAME'),
                        'geom': self._transform(feature.geom.geos),
                        'type': Place.Type.state
                    })

        for shapefile in shapefiles.county.get_shapefiles():
            ds = DataSource(shapefile)
            for feature in ds[0]:
                state_geoid = feature.get('GEOID')[:2]
                Place.objects.update_or_create(
                    geoid=feature.get('GEOID'),
                    defaults={
                        'name': feature.get('NAME'),
                        'geom': self._transform(feature.geom.geos),
                        'parent': Place.objects.get(geoid=state_geoid),
                        'type': Place.Type.county,
                    })

        for shapefile in shapefiles.city.get_shapefiles():
            ds = DataSource(shapefile)
            for feature in ds[0]:
                county_geoid = feature.get('GEOID')[:5]
                Place.objects.update_or_create(
                    geoid=feature.get('GEOID'),
                    defaults={
                        'name': feature.get('NAME'),
                        'geom': self._transform(feature.geom.geos),
                        'parent': Place.objects.get(geoid=county_geoid),
                        'type': Place.Type.city,
                    })

    def _transform(self, shape):
        shape = shape.simplify(tolerance=self.simplification, preserve_topology=True)
        if isinstance(shape, MultiPolygon):
            return shape
        elif isinstance(shape, Polygon):
            return MultiPolygon(shape)
        else:
            raise Exception()
